/* Agent Tcl
 * Bob Gray
 * 20 May 1996 
 * 
 * *** DIFFERENT THAN THE STANDARD TCL DISTRIBUTION ***
 *
 * This is the MODIFIED version of Tcl that is part of the Agent Tcl system.
 * This MODIFIED version allows the capture and restoration of a script's
 * internal state.
 * 
 * tclStackCmd.c --
 *
 *      All of the command routines -- e.g. Tcl_WhileCmd -- that
 *      must examine or change the execution stack have been moved into
 *      this file.  
 *
 * Copyright (c) 1995-1996 Bob Gray, Dartmouth College
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 *
 * See the file "agent.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifdef STACK

#include "tclInt.h"
#include "tclState.h"

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CaseCmd --
 *
 *	This procedure is invoked to process the "case" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

#define CMD_CaseBody	1

static int 
TclCaseBody(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    /* get the element on the top of the stack */

  Interp *iPtr           = (Interp *) interp;
  Stack  *stack          = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top];
 
    /* handle errors */

  if (iPtr -> resultCode == TCL_ERROR) {
    char msg[100];
    char *arm = get_state_variable (interp, "case_arm");
    sprintf (msg, "\n    (\"%.50s\" arm line %d)", arm, interp -> errorLine);
    Tcl_AddErrorInfo (interp, msg);
  }

    /* move on the rest of the script */

  element -> flag = CMD_REST_OF_SCRIPT;
  return (iPtr -> resultCode);
}

int 
Tcl_CaseCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    int i, result;
    int body;
    char *string;
    int caseArgc, splitArgs;
    char **caseArgv;

      /* get the element on the top of the stack */

    Interp *iPtr           = (Interp *) interp;
    Stack  *stack          = &iPtr -> execution_stack;
    Stack_Element *element = stack -> slots[stack -> top];

      /* have we just executed an arm of the "case" command? */

    if (element -> flag == CMD_CaseBody) {
      return TclCaseBody (dummy, interp, argc, argv);
    }

      /* or are we just starting the "case" command? */
 
    if (argc < 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " string ?in? patList body ... ?default body?\"",
		(char *) NULL);
	return TCL_ERROR;
    }
    string = argv[1];
    body = -1;
    if (strcmp(argv[2], "in") == 0) {
	i = 3;
    } else {
	i = 2;
    }
    caseArgc = argc - i;
    caseArgv = argv + i;

    /*
     * If all of the pattern/command pairs are lumped into a single
     * argument, split them out again.
     */

    splitArgs = 0;
    if (caseArgc == 1) {
	result = Tcl_SplitList(interp, caseArgv[0], &caseArgc, &caseArgv);
	if (result != TCL_OK) {
	    return result;
	}
	splitArgs = 1;
    }

    for (i = 0; i < caseArgc; i += 2) {
	int patArgc, j;
	char **patArgv;
	register char *p;

	if (i == (caseArgc-1)) {
	    interp->result = "extra case pattern with no body";
	    result = TCL_ERROR;
	    goto cleanup;
	}

	/*
	 * Check for special case of single pattern (no list) with
	 * no backslash sequences.
	 */

	for (p = caseArgv[i]; *p != 0; p++) {
	    if (isspace(UCHAR(*p)) || (*p == '\\')) {
		break;
	    }
	}
	if (*p == 0) {
	    if ((*caseArgv[i] == 'd')
		    && (strcmp(caseArgv[i], "default") == 0)) {
		body = i+1;
	    }
	    if (Tcl_StringMatch(string, caseArgv[i])) {
		body = i+1;
		goto match;
	    }
	    continue;
	}

	/*
	 * Break up pattern lists, then check each of the patterns
	 * in the list.
	 */

	result = Tcl_SplitList(interp, caseArgv[i], &patArgc, &patArgv);
	if (result != TCL_OK) {
	    goto cleanup;
	}
	for (j = 0; j < patArgc; j++) {
	    if (Tcl_StringMatch(string, patArgv[j])) {
		body = i+1;
		break;
	    }
	}
	ckfree((char *) patArgv);
	if (j < patArgc) {
	    break;
	}
    }

    match:
    if (body != -1) {
        char arm[100];
        sprintf (arm, "%.50s", caseArgv[body - 1]);
        set_state_variable (interp, "case_arm", arm);

        element -> flag        = CMD_CaseBody;
        element -> commandProc = TclCaseBody;
        TclPushStack (stack, CMD_REST_OF_SCRIPT, caseArgv[body], TCL_VOLATILE);

        result = TCL_OK;
	goto cleanup;
    }

    /*
     * Nothing matched:  return nothing.
     */

    result = TCL_OK;

    cleanup:
    if (splitArgs) {
	ckfree((char *) caseArgv);
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_CatchCmd --
 *
 *	This procedure is invoked to process the "catch" Tcl command.
 *      The procedure calls procedure Tcl_CatchResult
 *      to handle the result of the body.  See the user documentation for
 *      details on what "catch" does. 
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation. 
 *
 *----------------------------------------------------------------------
 */

#define CMD_CatchResult 1

static int 
TclCatchResult(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top]; 
  int result             = interp -> resultCode;

    /* move on to the rest of the script */
 
  element -> flag = CMD_REST_OF_SCRIPT;

    /* TCL_PERMIT and TCL_REWIND can not be caught */

  if ((result == TCL_PERMIT) || (result == TCL_REWIND)) {
    return (result);
  }

    /* everything else can be caught */

  if (argc == 3) {
    if (Tcl_SetVar(interp, argv[2], interp->result, 0) == NULL) {
      Tcl_SetResult(interp, "couldn't save command result in variable", TCL_STATIC);
      return TCL_ERROR;
    }
  }

  Tcl_ResetResult(interp);
  sprintf(interp -> result, "%d", result);
  return TCL_OK;
}

int 
Tcl_CatchCmd(dummy, interp, argc, argv)
    ClientData dummy;
    Tcl_Interp *interp;
    int argc;
    char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top]; 

    /* has the "catch" body been executed? */

  if (element -> flag == CMD_CatchResult) {
    return TclCatchResult (dummy, interp, argc, argv);
  }

    /* or are we just starting the "catch" command */
 
  if ((argc != 2) && (argc != 3)) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " command ?varName?\"", (char *) NULL);
    return TCL_ERROR;
  }

  element -> flag        = CMD_CatchResult;
  element -> commandProc = TclCatchResult;
  TclPushStack (stack, CMD_REST_OF_SCRIPT, argv[1], TCL_STATIC);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_EvalCmd --
 *
 *	This procedure is invoked to process the "eval" Tcl command.
 *      See the user documentation for details on what the "eval" command
 *      does.
 *
 * Results:
 *	A standard Tcl result.
 *
 *----------------------------------------------------------------------
 */

#define CMD_EvalResult 1

static int 
TclEvalResult(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top]; 

  if (iPtr -> resultCode == TCL_ERROR) {
    char msg[60];
    sprintf (msg, "\n    (\"eval\" body line %d)", interp -> errorLine);
    Tcl_AddErrorInfo (interp, msg);
  }

  element -> flag = CMD_REST_OF_SCRIPT;
  return (iPtr -> resultCode);
}

int 
Tcl_EvalCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top]; 

  if (element -> flag == CMD_EvalResult) {
    return TclEvalResult (dummy, interp, argc, argv);
  } else {
    char *cmd;
    Tcl_FreeProc *freeProc;

    if (argc < 2) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " arg ?arg ...?\"", (char *) NULL);
      return TCL_ERROR;
    }

    if (argc == 2) {
      cmd      = argv[1];
      freeProc = TCL_STATIC;
    } else {
      cmd      = Tcl_Concat (argc - 1, argv + 1);
      freeProc = TCL_DYNAMIC;
    }

    element -> flag = CMD_EvalResult;
    element -> commandProc = TclEvalResult;
    TclPushStack (stack, CMD_REST_OF_SCRIPT, cmd, freeProc);
    return TCL_OK;
  }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_ForCmd --
 *
 *	This procedure is invoked to process the "for" Tcl command.
 *      TclForTest evaluates the FOR loop test.  TclForBody executes
 *      the FOR loop body.  TclForReinit executes the FOR loop reinit.
 *      See the user documentation for details on the FOR command.
 *
 * Results:
 *	A standard Tcl result.
 *
 *----------------------------------------------------------------------
 */

#define CMD_ForInit 1 
#define CMD_ForBody 2  
#define CMD_ForReinit 3 

static int TclForInit _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp, int argc, char **argv));
static int TclForBody _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp, int argc, char **argv));
static int TclForReinit _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp, int argc, char **argv));

static int 
TclForExpression(interp, stack, element, expression, body)
     Tcl_Interp *interp;
     Stack *stack;
     Stack_Element *element;
     char *expression;
     char *body;
{
  int result;  /* error code from boolean expression evaluation */
  int value;   /* value of boolean expression                   */ 

  result = Tcl_ExprBoolean (interp, expression, &value);

  if (result != TCL_OK) {
    element -> flag = CMD_REST_OF_SCRIPT;
    return result;
  }

  if (!value) {
    element -> flag = CMD_REST_OF_SCRIPT;
    Tcl_ResetResult (interp);
    return TCL_OK;
  } else {
    element -> flag        = CMD_ForBody; 
    element -> commandProc = TclForBody;
    TclPushStack (stack, CMD_REST_OF_SCRIPT, body, TCL_STATIC); 
    return TCL_OK;
  }
}
 
static int 
TclForInit(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top]; 
  int result             = iPtr -> resultCode;

    /* see if the INIT command succeded */

  if (result != TCL_OK) {
    if (result == TCL_ERROR) {
       Tcl_AddErrorInfo(interp, "\n    (\"for\" initial command)");
    }

    return result;
  }

    /* evaluate the expression */

  return TclForExpression (interp, stack, element, argv[2], argv[4]);
}

static int 
TclForBody(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top]; 
  int result             = iPtr -> resultCode;

  if ((result != TCL_OK) && (result != TCL_CONTINUE)) {
     if (result == TCL_ERROR) {
       char msg[60];
       sprintf(msg, "\n    (\"for\" body line %d)", interp->errorLine);
       Tcl_AddErrorInfo(interp, msg);
     } else if (result == TCL_BREAK) {
       result = TCL_OK;
       Tcl_ResetResult(interp);
     }

     element -> flag = CMD_REST_OF_SCRIPT;
     return result;
  }

  element -> flag        = CMD_ForReinit;
  element -> commandProc = TclForReinit;
  TclPushStack (stack, CMD_REST_OF_SCRIPT, argv[3], TCL_STATIC);
  return TCL_OK;
}

static int 
TclForReinit(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top]; 
  int result             = iPtr -> resultCode;

  if ((result != TCL_OK) && (result != TCL_CONTINUE)) {
     if (result == TCL_ERROR) {
       char msg[60];
       sprintf(msg, "\n    (\"for\" loop-end command line %d)", interp->errorLine);
       Tcl_AddErrorInfo(interp, msg);
     } else if (result == TCL_BREAK) {
       result = TCL_OK;
       Tcl_ResetResult(interp);
     }

     element -> flag = CMD_REST_OF_SCRIPT;
     return result;
  }

    /* evaluate the test expression */

  return TclForExpression (interp, stack, element, argv[2], argv[4]);
}

int 
Tcl_ForCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top]; 

  if (element -> flag == CMD_ForBody) {
    return TclForBody (dummy, interp, argc, argv);
  } else if (element -> flag == CMD_ForReinit) {
    return TclForReinit (dummy, interp, argc, argv);
  } else if (element -> flag == CMD_ForInit) {
    return TclForInit (dummy, interp, argc, argv);
  } else if (argc != 5) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " start test next command\"", (char *) NULL);
    return TCL_ERROR;
  } else {
    element -> flag = CMD_ForInit;
    element -> commandProc = TclForInit;
    TclPushStack (stack, CMD_REST_OF_SCRIPT, argv[1], TCL_STATIC);
    return TCL_OK;
  }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_ForeachCmd --
 *
 *	This procedure is invoked to process the "foreach" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 *----------------------------------------------------------------------
 */

#define CMD_ForeachBody 1

#define STATIC_SIZE 4

typedef struct ForeachData {
    int j;				/* Current loop iteration            */
    int maxj; 				/* Number of loop iterations         */
    int numLists;			/* Number of value lists             */
    int indexArray[STATIC_SIZE];	/* Array of value list indices       */ 
    int varcListArray[STATIC_SIZE];     /* Number of loop variables per list */
    char **varvListArray[STATIC_SIZE];  /* Array of variable name lists      */
    int argcListArray[STATIC_SIZE];	/* Array of value list sizes         */
    char **argvListArray[STATIC_SIZE];  /* Array of value lists              */
    int *index;
    int *varcList;
    char ***varvList;
    int *argcList;
    char ***argvList;
} ForeachData;

static int 
TclForeachBody(clientData, interp, argc, argv)
     ClientData clientData;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    int i, v, k;
    Interp *iPtr;
    Stack *stack;
    Stack_Element *element;
    int result;
    ForeachData *data = (ForeachData *) clientData;
    char value_scratch[16];

	/* get the stack */

    iPtr = (Interp *) interp;
    stack = &iPtr -> execution_stack;
    element = stack -> slots[stack -> top];  
    result = iPtr -> resultCode;

	/* check the body result */

    if (result == TCL_BREAK) {

	result = TCL_OK;

    } else if (result == TCL_ERROR) {

	char msg[100];
	sprintf(msg, "\n    (\"foreach\" body line %d)", interp->errorLine);
	Tcl_AddErrorInfo(interp, msg);

    } else if ((result == TCL_OK) || (result == TCL_CONTINUE)) {

	result = TCL_OK;

	if (++data->j < data->maxj) { 

	    for (i=0 ; i < data->numLists ; i++) {
		for (v=0 ; v < data->varcList[i] ; v++) {
		    int k = data->index[i]++;
		    char *value = "";
		    if (k < data->argcList[i]) {
		        value = data->argvList[i][k];
		    }
		    if (Tcl_SetVar(interp, data->varvList[i][v], value, 0) == NULL) {
		        Tcl_AppendResult(interp, "couldn't set loop variable: \"",
			    data->varvList[i][v], "\"", (char *)NULL);
		        result = TCL_ERROR;
		    }
	        }
	    }

	    if (result == TCL_OK) {
      		integerToString (data->j, value_scratch); 
		set_state_variable (interp, "foreach_count", value_scratch);
		TclPushStack (stack, CMD_REST_OF_SCRIPT, argv[argc-1], TCL_STATIC); 
		return TCL_OK;
	    }
	}
    }

	/* done with Foreach so cleanup */

    if (result == TCL_OK) {
	Tcl_ResetResult (interp);
    }

    element -> flag = CMD_REST_OF_SCRIPT;

    for (i = 0 ; i < data->numLists ; i++) {
	if (data->argvList[i] != (char **)NULL) {
	    ckfree((char *) data->argvList[i]);
	}
	if (data->varvList[i] != (char **)NULL) {
	    ckfree((char *) data->varvList[i]);
	}
    }
    if (data->numLists > STATIC_SIZE) {
	ckfree((char *) data->index);
	ckfree((char *) data->varcList);
	ckfree((char *) data->argcList);
	ckfree((char *) data->varvList);
	ckfree((char *) data->argvList);
    }
    ckfree (data);
    return result;
}

int 
Tcl_ForeachCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    int i, j, v, k;
    int result = TCL_OK;
    Interp *iPtr;
    Stack *stack;
    Stack_Element *element;
    ForeachData *data;

	/* get the stack */

    iPtr = (Interp *) interp;
    stack = &iPtr -> execution_stack;
    element = stack -> slots[stack -> top];
 
	/* check the arguments */
 
    if ((argc < 4) || (argc % 2 != 0)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" varList list ?varList list ...? command\"", (char *) NULL);
	return TCL_ERROR;
    }

	/* allocate the ForeachData structure */

    data = (ForeachData *) ckalloc ((unsigned) sizeof(ForeachData));
    element->commandData = (ClientData) data;

	/*
	 * Manage numList parallel value lists.
	 * argvList[i] is a value list counted by argcList[i]
	 * varvList[i] is the list of variables associated with the value list
	 * varcList[i] is the number of variables associated with the value list
	 * index[i] is the current pointer into the value list argvList[i]
	 */

    data->numLists = (argc - 2) / 2;

    if (data->numLists > STATIC_SIZE) {
	data->index = (int *) ckalloc(data->numLists * sizeof(int));
	data->varcList = (int *) ckalloc(data->numLists * sizeof(int));
	data->varvList = (char ***) ckalloc(data->numLists * sizeof(char **));
	data->argcList = (int *) ckalloc(data->numLists * sizeof(int));
	data->argvList = (char ***) ckalloc(data->numLists * sizeof(char **));
    } else {
	data->index = data->indexArray;
	data->varcList = data->varcListArray; 
	data->varvList = data->varvListArray; 
	data->argcList = data->argcListArray; 
	data->argvList = data->argvListArray;
    } 
    
    for (i = 0 ; i < data->numLists ; i++) {
	data->index[i] = 0;
	data->varcList[i] = 0;
	data->varvList[i] = (char **)NULL;
	data->argcList[i] = 0;
	data->argvList[i] = (char **)NULL;
    }

	/*
	 * Break up the value lists and variable lists into elements
	 */

    data->maxj = 0;

    for (i = 0; i < data->numLists ; i++) {
	result = Tcl_SplitList(interp, argv[1+i*2], &data->varcList[i], &data->varvList[i]);
	if (result != TCL_OK) {
	    goto errorReturn;
	}
	result = Tcl_SplitList(interp, argv[2+i*2], &data->argcList[i], &data->argvList[i]);
	if (result != TCL_OK) {
	    goto errorReturn;
	}
	j = data->argcList[i] / data->varcList[i];
	if ((data->argcList[i] % data->varcList[i]) != 0) {
	    j++;
	}
	if (j > data->maxj) {
	    data->maxj = j;
	}
    }

	/* done if we have no elements */

    if (data->maxj == 0) {
	goto errorReturn;
    }

	/* pick the correct list element */

    if (element -> flag == CMD_ForeachBody) {

	Tcl_GetInt (interp, get_state_variable(interp, "foreach_count"), &data->j);

	for (j = 0; j < data->j; j++) {
	    for (i = 0; i < data->numLists; i++) {
		for (v = 0; v < data->varcList[i]; v++) {
		    data->index[i] += 1;
		}
	    }
	}

    } else {

	data->j = 0;
   }
	    
	/* set the loop control variable(s) */

   for (i=0 ; i < data->numLists ; i++) {
	for (v=0 ; v < data->varcList[i] ; v++) {
	    char *value = "";
	    int k = data->index[i]++;
	    if (k < data->argcList[i]) {
		value = data->argvList[i][k];
	    }
	    if (Tcl_SetVar(interp, data->varvList[i][v], value, 0) == NULL) {
		Tcl_AppendResult(interp, "couldn't set loop variable: \"",
		    data->varvList[i][v], "\"", (char *)NULL);
		result = TCL_ERROR;
		goto errorReturn;
	    }
	}
    }

	/* either push body or check body result */

    if (element -> flag == CMD_ForeachBody) {
	return (TclForeachBody ((ClientData) data, interp, argc, argv));
    } else {
	element -> flag        = CMD_ForeachBody;
	element -> commandProc = TclForeachBody;
	set_state_variable (interp, "foreach_count", "0");
	TclPushStack (stack, CMD_REST_OF_SCRIPT, argv[argc-1], TCL_STATIC);
	return TCL_OK;
    }

	/* non-normal return */

errorReturn:
    for (i = 0 ; i < data->numLists ; i++) {
	if (data->argvList[i] != (char **)NULL) {
	    ckfree((char *) data->argvList[i]);
	}
	if (data->varvList[i] != (char **)NULL) {
	    ckfree((char *) data->varvList[i]);
	}
    }
    if (data->numLists > STATIC_SIZE) {
	ckfree((char *) data->index);
	ckfree((char *) data->varcList);
	ckfree((char *) data->argcList);
	ckfree((char *) data->varvList);
	ckfree((char *) data->argvList);
    }
    return result;
}

#undef STATIC_SIZE

/*
 *----------------------------------------------------------------------
 *
 * Tcl_IfCmd --
 *
 *	This procedure is invoked to process the "if" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 *----------------------------------------------------------------------
 */

int 
Tcl_IfCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  int i = 1;
  int result;      /* error code from boolean expression evaluation */
  int value;       /* value of boolean expression                   */
  Interp *iPtr = (Interp *) interp;
  Stack *stack = &iPtr -> execution_stack;

  while (1) {

    /*
     * At this point in the loop, argv and argc refer to an expression
     * to test, either for the main expression or an expression
     * following an "elseif".  The arguments after the expression must
     * be "then" (optional) and a script to execute if the expression is
     * true.
     */

    if (i >= argc) {
      Tcl_AppendResult(interp, "wrong # args: no expression after \"", argv[i-1], "\" argument", (char *) NULL);
      return TCL_ERROR;
    }

    if ((result = Tcl_ExprBoolean(interp, argv[i], &value)) != TCL_OK) {
      return result;
    }

    i++;
    if ((i < argc) && (strcmp(argv[i], "then") == 0)) {
      i++;
    }

    if (i >= argc) {
      Tcl_AppendResult(interp, "wrong # args: no script following \"", argv[i-1], "\" argument", (char *) NULL);
      return TCL_ERROR;
    }

    if (value) {
      TclPushStack (stack, CMD_REST_OF_SCRIPT, argv[i], TCL_STATIC);
      return TCL_OK;
    }

    /*
     * The expression evaluated to false.  Skip the command, then
     * see if there is an "else" or "elseif" clause.
     */

    if (++i >= argc) { 
      return TCL_OK;
    }

    if ((argv[i][0] == 'e') && (strcmp(argv[i], "elseif") == 0)) {
      i++;
      continue;
    }

    break;
  }

  /*
   * Couldn't find a "then" or "elseif" clause to execute.  Check now
   * for an "else" clause.  We know that there's at least one more
   * argument when we get here.
   */

  if (!strcmp(argv[i], "else")) {
    if (++i >= argc) {
      Tcl_AppendResult(interp, "wrong # args: no script following \"else\" argument", (char *) NULL);
      return TCL_ERROR;
    }
  }

  TclPushStack (stack, CMD_REST_OF_SCRIPT, argv[i], TCL_STATIC);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SwitchCmd --
 *
 *	This procedure is invoked to process the "switch" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

#define CMD_SwitchBody	1

#define EXACT		0
#define GLOB		1
#define REGEXP		2

static int 
TclSwitchBody(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    /* get the top element of the stack */

  Interp        *iPtr    = (Interp *) interp;
  Stack         *stack   = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top];

    /* check for an error message */

  if (iPtr -> resultCode == TCL_ERROR) {
    char msg[100];
    char *arm = get_state_variable (interp, "switch_arm");
    sprintf (msg, "\n    (\"%s\" arm line %d)", arm, interp -> errorLine);
    Tcl_AddErrorInfo (interp, msg);
  }
  
    /* move on the rest of the script */

  element -> flag = CMD_REST_OF_SCRIPT;
  return (iPtr -> resultCode);  
}

int 
Tcl_SwitchCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    int i, code, mode, matched;
    int body;
    char *string;
    int switchArgc, splitArgs;
    char **switchArgv;

      /* get the top element of the stack */

    Interp        *iPtr    = (Interp *) interp;
    Stack         *stack   = &iPtr -> execution_stack;
    Stack_Element *element = stack -> slots[stack -> top];

      /* have we just executed an arm of the "switch" command? */

    if (element -> flag == CMD_SwitchBody) {
      return TclSwitchBody (dummy, interp, argc, argv);
    }

      /* or are we just starting the command? */

    switchArgc = argc-1;
    switchArgv = argv+1;
    mode = EXACT;
    while ((switchArgc > 0) && (*switchArgv[0] == '-')) {
	if (strcmp(*switchArgv, "-exact") == 0) {
	    mode = EXACT;
	} else if (strcmp(*switchArgv, "-glob") == 0) {
	    mode = GLOB;
	} else if (strcmp(*switchArgv, "-regexp") == 0) {
	    mode = REGEXP;
	} else if (strcmp(*switchArgv, "--") == 0) {
	    switchArgc--;
	    switchArgv++;
	    break;
	} else {
	    Tcl_AppendResult(interp, "bad option \"", switchArgv[0],
		    "\": should be -exact, -glob, -regexp, or --",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	switchArgc--;
	switchArgv++;
    }
    if (switchArgc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " ?switches? string pattern body ... ?default body?\"",
		(char *) NULL);
	return TCL_ERROR;
    }
    string = *switchArgv;
    switchArgc--;
    switchArgv++;

    /*
     * If all of the pattern/command pairs are lumped into a single
     * argument, split them out again.
     */

    splitArgs = 0;
    if (switchArgc == 1) {
	code = Tcl_SplitList(interp, switchArgv[0], &switchArgc, &switchArgv);
	if (code != TCL_OK) {
	    return code;
	}
	splitArgs = 1;
    }

    for (i = 0; i < switchArgc; i += 2) {
	if (i == (switchArgc-1)) {
	    interp->result = "extra switch pattern with no body";
            code = TCL_ERROR;
            goto cleanup; 
	}

      /*
       * See if the pattern matches the string.
       */

      matched = 0;
      if ((*switchArgv[i] == 'd') && (i == switchArgc-2)
		&& (strcmp(switchArgv[i], "default") == 0)) {
	matched = 1;
      } else {
        switch (mode) {
    	  case EXACT:
	     matched = (strcmp(string, switchArgv[i]) == 0);
	     break;
	  case GLOB:
	     matched = Tcl_StringMatch(string, switchArgv[i]);
	     break;
  	  case REGEXP:
	     matched = Tcl_RegExpMatch(interp, string, switchArgv[i]);
	     if (matched < 0) {
	       code = TCL_ERROR;
               goto cleanup;
             }
	     break;
	}
      }
      if (!matched) {
        continue;
      }

	/*
	 * We've got a match.  Find a body to execute, skipping bodies
	 * that are "-".
	 */

	for (body = i + 1; ; body += 2) {
	  if (body >= switchArgc) {
	     Tcl_AppendResult(interp, "no body specified for pattern \"", switchArgv[i], "\"", (char *) NULL);
             code = TCL_ERROR;
	     goto cleanup;
	  }

	  if ((switchArgv[body][0] != '-') || (switchArgv[body][1] != 0)) {

	     char arm[100];
             sprintf (arm, "%.50s", switchArgv[i]);
	     set_state_variable (interp, "switch_arm", arm);

             element -> flag = CMD_SwitchBody;
             element -> commandProc = TclSwitchBody; 
             TclPushStack (stack, CMD_REST_OF_SCRIPT, switchArgv[body], TCL_VOLATILE);

             code = TCL_OK;
             goto cleanup;
          } 
	}
    }

    /*
     * Nothing macthed:  return nothing
     */

    code = TCL_OK;

    cleanup:
    if (splitArgs) {
      ckfree ((char *) switchArgv);
    }
    return code;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_WhileCmd --
 *
 *	This procedure is invoked to process the "while" Tcl command.
 *      Procedure TclWhileTest evaluates the control expression.
 *      Procedure TclWhileBody evaluates the body.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 *----------------------------------------------------------------------
 */

#define CMD_WhileBody 1

static int TclWhileBody _ANSI_ARGS_((ClientData dummy, Tcl_Interp *interp, int argc, char **argv));

static int 
TclWhileExpression(interp, stack, element, expression, body)
     Tcl_Interp *interp;
     Stack *stack;
     Stack_Element *element;
     char *expression;
     char *body;
{
  int result;   /* error code from boolean expression evaluation */
  int value;    /* value of boolean expression                   */

  result = Tcl_ExprBoolean(interp, expression, &value);

  if (result != TCL_OK) {
    element -> flag = CMD_REST_OF_SCRIPT;
    return result;
  }

  if (!value) {
    element -> flag = CMD_REST_OF_SCRIPT;
    Tcl_ResetResult (interp);
    return TCL_OK;
  } else {
    element -> flag        = CMD_WhileBody;
    element -> commandProc = TclWhileBody;
    TclPushStack (stack, CMD_REST_OF_SCRIPT, body, TCL_STATIC);
    return TCL_OK;
  }
}   

static int 
TclWhileBody(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top];   
  int result             = iPtr -> resultCode;
   
  if ((result != TCL_OK) && (result != TCL_CONTINUE)) {
     if (result == TCL_ERROR) {
       char msg[60];
       sprintf(msg, "\n    (\"while\" body line %d)", interp->errorLine);
       Tcl_AddErrorInfo(interp, msg);
     } else if (result == TCL_BREAK) {
       result = TCL_OK;
       Tcl_ResetResult(interp);
     }

     element -> flag = CMD_REST_OF_SCRIPT;
     return result;
  }

  /* evaluate the test expression */

  return TclWhileExpression (interp, stack, element, argv[1], argv[2]);
}

int 
Tcl_WhileCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
    Interp *iPtr           = (Interp *) interp;
    Stack *stack           = &iPtr -> execution_stack;
    Stack_Element *element = stack -> slots[stack -> top];   
    
    if (element -> flag == CMD_WhileBody) {
      return TclWhileBody (dummy, interp, argc, argv);
    } else if (argc != 3) {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " test command\"", (char *) NULL);
      return TCL_ERROR;
    } else {
      return TclWhileExpression (interp, stack, element, argv[1], argv[2]);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SourceCmd --
 *
 *	This procedure is invoked to process the "source" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

#define CMD_SourceBody 1
 
static int 
TclSourceBody(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  Interp *iPtr           = (Interp *) interp;
  Stack *stack           = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top];
  int result             = iPtr -> resultCode;
  char *parentScript;

    /* handle errors */

  if (result == TCL_RETURN) {
    result = TclUpdateReturnInfo (iPtr);
  } else if (result == TCL_ERROR) {
    char msg[200];
    sprintf (msg, "\n    (file \"%.150s\" line %d)", argv[1], interp -> errorLine);
    Tcl_AddErrorInfo (interp, msg);
  }

    /* restore the name of the parent script */

  ckfree ((char *) iPtr -> scriptFile);

  if ((parentScript = get_state_variable (interp, "parent_script")) == NULL) {
    iPtr -> scriptFile = NULL;
  } else {
    iPtr -> scriptFile = ckalloc ((long) strlen(parentScript) + 1);
    strcpy (iPtr -> scriptFile, parentScript);
  }  

    /* move on to the rest of the script */
  
  element -> flag = CMD_REST_OF_SCRIPT;
  return result; 
}

int 
Tcl_SourceCmd(dummy, interp, argc, argv)
     ClientData dummy;
     Tcl_Interp *interp;
     int argc;
     char **argv;
{
  char *cmdBuffer;
  Tcl_DString buffer;  
  Interp *iPtr = (Interp *) interp;
  Stack *stack = &iPtr -> execution_stack;
  Stack_Element *element = stack -> slots[stack -> top];

    /* have we already sourced the file? */

  if (element -> flag == CMD_SourceBody) {
    return TclSourceBody (dummy, interp, argc, argv);
  }

    /* or are we just starting */

  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " fileName\"", (char *) NULL);
    return TCL_ERROR;
  }

    /* load the file */

  if ((cmdBuffer = Tcl_LoadScript (interp, argv[1], &buffer)) == NULL) {
    return TCL_ERROR;
  }

    /* save the old script file and set the new one */

  if (iPtr -> scriptFile != NULL) {
    if (set_state_variable (interp, "parent_script", iPtr -> scriptFile) == NULL) {
      Tcl_AppendResult (interp, "unable to save the name of the parent script", (char *) NULL);
      Tcl_DStringFree (&buffer);
      return TCL_ERROR;
    }
  }

  iPtr -> scriptFile = ckalloc ((long) strlen(argv[1]) + 1);
  strcpy (iPtr -> scriptFile, argv[1]);

    /* push file contents onto the stack */
 
  element -> flag        = CMD_SourceBody;
  element -> commandProc = TclSourceBody;
  TclPushStack (stack, CMD_REST_OF_SCRIPT, cmdBuffer, TCL_DYNAMIC);
  Tcl_DStringFree (&buffer);
  return TCL_OK;        
}

#endif    /* STACK */
