/* Agent Tcl
 * Bob Gray
 * 20 May 1996 
 * 
 * *** DIFFERENT THAN THE STANDARD TCL DISTRIBUTION *** 
 *
 * This is the MODIFIED version of Tcl that is part of the Agent Tcl system.
 * This MODIFIED version allows the capture and restoration of a script's
 * internal state.
 * 
 * tclStackRun.c --
 *
 *      This file contains the new version of Tcl_Eval which is
 *      based on an explit stack rather than recursive calls.
 *
 * Copyright (c) 1995-1996 Bob Gray, Dartmouth College
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 *
 * See the file "agent.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifdef STACK
 
#include "tclInt.h"
#ifndef TCL_GENERIC_ONLY
#include "tclPort.h"
#endif

#define NUM_CHARS 200

static char copyStorage[NUM_CHARS];

static Command *TclFindCommand 
  _ANSI_ARGS_((Interp *iPtr, Stack *stack, Stack_Element *element));

static void TclEvalError 
  _ANSI_ARGS_((Interp *iPtr, Stack_Element *stack_command, char *ellipsis));

/*
 *-----------------------------------------------------------------
 *
 *  TclEvalError ---
 *
 *     Record information about the errors that occur during Tcl_Eval
 *
 *-----------------------------------------------------------------
 */

static void 
TclEvalError(iPtr, stack_command, ellipsis)
     Interp *iPtr;
     Stack_Element *stack_command;
     char *ellipsis;
{
    Tcl_Interp *interp = (Tcl_Interp *) iPtr; 
    int result         = iPtr -> resultCode;

	/* handle errors that occur on the top stack command */

    if (iPtr -> execution_stack.top == 0) {
	if (result == TCL_RETURN) {
	    result = TclUpdateReturnInfo (iPtr);
	}
	if ((result != TCL_OK) && (result != TCL_REWIND) && (result != TCL_PERMIT) && (result != TCL_ERROR) && !(stack_command -> evalFlags & TCL_ALLOW_EXCEPTIONS)) {
	    Tcl_ResetResult (interp);
	    if (result == TCL_BREAK) { 
		iPtr -> result = "invoked \"break\" outside of a loop";
	    } else if (result == TCL_CONTINUE) {
		iPtr -> result = "invoked \"continue\" outside of a loop";
	    } else {
		iPtr -> result = iPtr -> resultSpace;
		sprintf(iPtr -> resultSpace, "command returned bad code: %d", result);
	    }
	    result = TCL_ERROR;
	}
    }

    /* record information about what was being executed when the error */
    /* occurred if there was an error                                  */

  if (((result == TCL_PERMIT) || (result == TCL_ERROR)) && !(iPtr -> flags & ERR_ALREADY_LOGGED) && !(stack_command -> evalFlags & TCL_KNOWN_PROC)) {

    int numChars;
    register char *p;

      /* compute the line number where the error occurred */

    iPtr -> errorLine = 1;

    for (p = stack_command -> cmd; p != stack_command -> cmdStart; p++)
      if (*p == '\n') 
	iPtr -> errorLine++;

      /* figure out how much of the command to print in the error message  */
      /* (up to a certain number of characters or up to the first newline) */

    numChars = stack_command -> rest - stack_command -> cmdStart;

    if (numChars > NUM_CHARS - 50) {
      numChars = NUM_CHARS-50;
      ellipsis = " ...";
    }

    if (!(iPtr->flags & ERR_IN_PROGRESS)) {
      sprintf(copyStorage, "\n    while executing\n\"%.*s%s\"", numChars, stack_command -> cmdStart, ellipsis);
    } else {
      sprintf(copyStorage, "\n    invoked from within\n\"%.*s%s\"", numChars, stack_command -> cmdStart, ellipsis);
    }
	
    Tcl_AddErrorInfo(interp, copyStorage);
    iPtr->flags &= ~ERR_ALREADY_LOGGED;

  } else {

    iPtr -> flags &= ~ERR_ALREADY_LOGGED;
  }

  iPtr -> resultCode = result;
}

/*
 *-----------------------------------------------------------------
 *
 * Tcl_Eval --
 *
 *	Parse and execute a command in the Tcl language.
 *
 * Results:
 *	The return value is one of the return codes defined in tcl.hd
 *	(such as TCL_OK), and interp->result contains a string value
 *	to supplement the return code.  The value of interp->result
 *	will persist only until the next call to Tcl_Eval:  copy it or
 *	lose it! *TermPtr is filled in with the character just after
 *	the last one that was part of the command (usually a NULL
 *	character or a closing bracket).
 *
 *-----------------------------------------------------------------
 */

static Command *
TclFindCommand(iPtr, stack, element)
	Interp *iPtr;
	Stack *stack;
	Stack_Element *element;
{
  Tcl_Interp *interp = (Tcl_Interp *) iPtr;
  Tcl_HashEntry *hPtr;
  Command *command;
  int i;

    /* find the command */

  hPtr = Tcl_FindHashEntry(&iPtr -> commandTable, element -> argv[0]);
 
    /* find the "unknown" command if the command was not found */

  if (hPtr == NULL) {
    if ((hPtr = Tcl_FindHashEntry (&iPtr -> commandTable, "unknown")) == NULL) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, "invalid command name \"", element -> argv[0], "\"", (char *) NULL);
      return NULL;
    }

    for (i = element -> argc; i >= 0; i--) {
      element -> argv[i+1] = element -> argv[i];
    }

    element -> argv[0] = "unknown";
    element -> argc++;
  }

    /* set the command flags */

  command = (Command *) Tcl_GetHashValue (hPtr);

  if (stack -> top == 0) {
    element->commandFlags = command->flags;
  } else {
    Stack_Element *previous = stack->slots[stack->top - 1];
    element->commandFlags = (command->flags) & (previous->commandFlags);
  }

  return (command);
}

int 
Tcl_EvalStack (interp, place)
     Tcl_Interp *interp;
     int place;
{
    int length;
    int i;                  /* loop counter                                    */
    int newTop;             /* old top of the stack                            */
    Command *cmdPtr;        /* command information                             */
    Trace *tracePtr;        /* command trace information                       */
    char *oldBuffer;        /* old buffer space                                */ 
    register char *src;     /* current character in command                    */
    register char c;        /* current character in command                    */
    char *termPtr;          /* character after last one in command             */
    int argSize;            /* current number of slots in the argument list    */ 
    int result;             /* result code                                     */
    Tcl_HashEntry *hPtr;

	/* stack */
 
    Interp *iPtr;
    register Stack *stack;
    register Stack_Element *stack_command;
 
	/* get the stack */

    iPtr = (Interp *) interp;
    stack = &iPtr -> execution_stack;
    
	/* push the command onto the stack OR run until stack empty */

    if (place == STACK_BOTTOM) {
	newTop = 0;
    } else {
	newTop = stack -> top;
    }

	/* outer loop iterates until we are finished with the command */

    while (stack -> top >= newTop) {

	    /* get the current command */

	stack_command = stack -> slots[stack -> top];
	src           = stack_command -> rest;
	termPtr	      = src;

	    /* are we in the middle of a command substitution?  OR  */
	    /* is the flag still CMD_START which means that we have */
	    /* just finished a command that ignores the stack? OR   */
	    /* is it an already parsed command which means that we  */
	    /* can go ahead and execute it?			    */
 
	if (stack_command -> subArgc != 0) {
	    length = strlen (iPtr -> result);
	    stack_command->subArgv[stack_command->subArgc] = ckalloc ((unsigned) length + 1);
	    strcpy (stack_command->subArgv[stack_command->subArgc], iPtr -> result);
	    stack_command -> subArgc = 0; 
	} else if (stack_command -> flag == CMD_START) {
	    stack_command -> flag = CMD_REST_OF_SCRIPT;
	} else if (stack_command -> flag == CMD_KNOWN_PROC) {
	    stack_command -> flag = CMD_START;
	} 

	    /* move on to the rest of the script or execute current command */

	if (stack_command -> flag == CMD_REST_OF_SCRIPT) {

		/* move up the stack if there was an error */

	    if (iPtr -> resultCode != TCL_OK) {
		TclEvalError (iPtr, stack_command, "");
		TclPopStack (stack);
		continue;
            }

		/* skim off leading white space and comments */

	    while (1) {

		c = *src;

		while ((CHAR_TYPE(c) == TCL_SPACE) || (c == ';') || (c == '\n')) {	  
		    src++;
		    c = *src;
		}
      
		if (c != '#') {
		    break;
		}

		while (*src != 0) {
		    if (*src == '\\') {
			int length;
			Tcl_Backslash(src, &length);
			src += length;
		    } else if (*src == '\n') {
			src++;	
			termPtr = src;
			break;
		    } else {
			src++;
		    }
		}          
	    }	

		/*  move up stack if we have reached terminator */  

	    if (*src == stack_command -> termChar) {
		TclPopStack (stack);
		continue;
	    }
 
		/* bail if we have deleted the interpreter */

	    if (iPtr -> flags & DELETED) {
		Tcl_ResetResult ((Tcl_Interp *) iPtr);
		iPtr -> result = "attempt to call eval in deleted interpreter";
		iPtr -> resultCode = TCL_ERROR;
		Tcl_SetErrorCode ((Tcl_Interp *) iPtr, "CORE", "IDELETE", iPtr -> result, (char *) NULL);
		TclPopStack (stack);
		continue;
	    }

		/* bail if we have too many nested calls */

	    if (stack -> top > iPtr -> maxNestingDepth) {
		Tcl_ResetResult ((Tcl_Interp *) iPtr);
		iPtr -> result = "too many commands on the stack (infinite recursion?)";
		iPtr -> resultCode = TCL_ERROR;
		TclPopStack (stack);
		continue;
	    }

		/* update the stack pointers */

	    stack_command -> cmdStart = src;
	    stack_command -> rest     = src;

		/* parse the words of the command (generate argc and argv) */

	    argSize                   = ST_ARGS;
	    oldBuffer                 = stack_command -> pv.buffer;
	    stack_command -> pv.next  = stack_command -> pv.buffer;
	    stack_command -> argc     = 0;

	    if (stack -> top == 0) {
		stack_command -> commandFlags = TCL_INTERRUPT;
	    } else {
		Stack_Element *previous = stack -> slots[stack -> top - 1];
		stack_command -> commandFlags = previous -> commandFlags;
	    }

	    if (stack_command -> evalFlags & TCL_NO_STATE) {
		stack_command -> commandFlags &= ~TCL_INTERRUPT;
	    }

	    while (1) {

		int newArgs, maxArgs;
		char **newArgv;
		int i;

		    /*
		     * Note:  the "- 2" below guarantees that we won't use the
		     * last two argv slots here.  One is for a NULL pointer to
		     * mark the end of the list, and the other is to leave room
		     * for inserting the command name "unknown" as the first
		     * argument (see below).
		     */

		maxArgs = argSize - stack_command -> argc - 2;
  
		iPtr -> resultCode = 
		    TclParseWords((Tcl_Interp *) iPtr, src, 
                    stack_command -> evalFlags,
		    maxArgs, &termPtr, &newArgs, 
                    &stack_command -> argv[stack_command -> argc], 
                    &stack_command -> pv);

		src = termPtr;

		if (iPtr -> resultCode != TCL_OK) {
		    break;
		}

		    /*
		     * Careful!  Buffer space may have gotten reallocated while
		     * parsing words.  If this happened, be sure to update all
		     * of the older argv pointers to refer to the new space.
		     */

		if (oldBuffer != stack_command -> pv.buffer) {
		    int i;
		    for (i = 0; i < stack_command -> argc; i++) {
			stack_command -> argv[i] += 
			(stack_command -> pv.buffer - oldBuffer);
		    }
		    oldBuffer = stack_command -> pv.buffer;
		}
		stack_command -> argc += newArgs;
		if (newArgs < maxArgs) {
		    break;
		}

		    /*
		     * Args didn't all fit in the current array.  Make it bigger.
		     */

		argSize *= 2;
		newArgv = (char **)
		    ckalloc((unsigned) argSize * sizeof(char *));
		for (i = 0; i < stack_command -> argc; i++) {
		    newArgv[i] = stack_command -> argv[i];
		}
		if (stack_command -> argv != stack_command -> argStore) {
		    ckfree((char *) stack_command -> argv);
		}
		stack_command -> argv = newArgv;
	    }

		/* unset the state variables associated with the substitutions */

	    if (stack_command -> subArgc != 0) {

		for (i = 1; i <= stack_command -> subArgc; i++) {
		    if (stack_command -> subArgv[i] != NULL) {
			ckfree ((char *) stack_command -> subArgv[i]);
			stack_command -> subArgv[i] = (char *) NULL;
		    }
		}

		stack_command -> subArgc = 0; 
	    }

	    stack_command -> rest = src;

		/* move up stack on error */

	    if (iPtr -> resultCode != TCL_OK) {
		TclEvalError (iPtr, stack_command, "...");
		TclPopStack (stack);
		continue;
	    }

		/* move to rest of command on empty command */
 
	    if ((stack_command -> argc == 0) || iPtr -> noEval) {
		continue;
	    }

	    stack_command -> argv[stack_command -> argc] = NULL;

		/* record history if desired */

	    if (stack_command -> evalFlags & TCL_RECORD_BOUNDS) {
		iPtr -> evalFirst = stack_command -> cmdStart;
		iPtr -> evalLast  = stack_command -> rest - 1;
	    }
 
		/* find the procedure to execute this command */

	    hPtr = Tcl_FindHashEntry(&iPtr -> commandTable, stack_command -> argv[0]);
 
	    if (hPtr == NULL) {
		if ((hPtr = Tcl_FindHashEntry (&iPtr -> commandTable, "unknown")) == NULL) {
		    Tcl_ResetResult(interp);
		    Tcl_AppendResult(interp, "invalid command name \"", stack_command -> argv[0], "\"", (char *) NULL);
	            iPtr -> resultCode = TCL_ERROR;
	            TclEvalError (iPtr, stack_command, "");
	            TclPopStack (stack); 
	            continue;
		}

	        for (i = stack_command -> argc; i >= 0; i--) {
		    stack_command -> argv[i+1] = stack_command -> argv[i];
	        }

	        stack_command -> argv[0] = "unknown";
	        stack_command -> argc++;
	    }

	    cmdPtr = (Command *) Tcl_GetHashValue (hPtr);

	    if (stack -> top == 0) {
	        stack_command -> commandFlags = cmdPtr -> flags;
	    } else {
	        Stack_Element *previous = stack->slots[stack->top - 1];
	        stack_command -> commandFlags = (cmdPtr -> flags) & (previous -> commandFlags);
	    }

	    if (stack_command -> evalFlags & TCL_NO_STATE) {
		stack_command -> commandFlags &= ~TCL_INTERRUPT;
	    }

	    stack_command -> flag        = CMD_START;
	    stack_command -> commandProc = cmdPtr -> proc;
	    stack_command -> commandData = cmdPtr -> clientData;

		/* call trace procedures */

	    for (tracePtr = iPtr->tracePtr; tracePtr != NULL; tracePtr = tracePtr->nextPtr) {
		char saved;

		if ((tracePtr -> level < iPtr -> execution_stack.top) && (tracePtr -> level != -1)) {
		    continue;
		}
		saved = *src;
		*src = 0;

		(*tracePtr->proc)(tracePtr->clientData, interp, 
		    iPtr->execution_stack.top,
		    stack_command -> cmdStart,
		    stack_command -> commandProc, stack_command -> commandData,
		    stack_command -> argc, stack_command -> argv);

		*src = saved;
	    }

		/* make sure that we have an empty result */

	    Tcl_ResetResult((Tcl_Interp *) iPtr);
	    iPtr->cmdCount++;

	} else if (stack_command -> commandProc == NULL) {

	    if ((cmdPtr = TclFindCommand (iPtr, stack, stack_command)) == NULL) {
		iPtr -> resultCode = TCL_ERROR;
		TclPopStack (stack); 
		continue;
	    }         

	    stack_command -> commandProc = cmdPtr -> proc;
	    stack_command -> commandData = cmdPtr -> clientData;
	}
         
	    /* invoke the command procedure */

	iPtr -> resultCode = (*stack_command->commandProc) (stack_command->commandData, interp, stack_command -> argc, stack_command -> argv);

	    /* unset the state variables associated with substitutions */

	if (stack_command -> subArgc != 0) {

	    for (i = 1; i <= stack_command -> subArgc; i++) {
		if (stack_command -> subArgv[i] != NULL) {
		     ckfree ((char *) stack_command -> subArgv[i]);
		     stack_command -> subArgv[i] = (char *) NULL;
	        }
	    }

	    stack_command -> subArgc = 0; 
	}

             /* handle asynchronous handlers */

	if (Tcl_AsyncReady()) {
	    iPtr -> evalFlags = TCL_NO_STATE;
	    iPtr -> resultCode = Tcl_AsyncInvoke(interp, iPtr -> resultCode);
	    iPtr -> evalFlags = 0;
	}
	if (iPtr -> resultCode != TCL_OK) {
	    TclEvalError (iPtr, stack_command, "");
	    TclPopStack (stack);
	}
    } 

	/* done */ 

    iPtr -> termPtr = src;
    return (iPtr -> resultCode);
}

int
Tcl_Eval (interp,cmd)
    Tcl_Interp *interp;
    char *cmd;
{
    int code;
    Interp *iPtr;
    Stack *stack;
    Stack_Element *command;

	/* get the stack */

    iPtr = (Interp *) interp;
    stack = &iPtr -> execution_stack;

	/* push the command onto the stack */

    TclPushStack (stack, CMD_REST_OF_SCRIPT, cmd, TCL_STATIC);

	/* set the flags */

    command               = stack -> slots[stack -> top];
    command -> evalFlags |= iPtr -> evalFlags;
    iPtr -> evalFlags     = 0;

    if (!(command -> evalFlags & TCL_STATE)) {
	command -> evalFlags &= ~TCL_STATE;
	command -> evalFlags |= TCL_NO_STATE;
    }

    if (command -> evalFlags & TCL_BRACKET_TERM) {
	command -> termChar = ']';
    }

	/* empty the result and evaluate */

    Tcl_ResetResult ((Tcl_Interp *) iPtr);
    iPtr->resultCode = TCL_OK;
    code = Tcl_EvalStack (interp, STACK_CURRENT);
    return (code);
}

int
Tcl_EvalKnownProc (interp,proc,data,argc,argv)
    Tcl_Interp *interp;
    Tcl_CmdProc proc;
    ClientData data;
    int argc;
    char **argv;
{
    int code;
    Interp *iPtr;
    Stack *stack;
    Stack_Element *command;

	/* get the stack */

    iPtr = (Interp *) interp;
    stack = &iPtr -> execution_stack;

	/* push on the command */

    TclPushKnown (stack, proc, data, argc, argv);

	/* set the flags -- be sure to OR */

    command               = stack -> slots[stack -> top];
    command -> evalFlags |= iPtr -> evalFlags;
    command -> evalFlags &= ~TCL_STATE;
    command -> evalFlags |= TCL_NO_STATE;
    iPtr -> evalFlags     = 0;

	/* empty out the result and evaluate */

    Tcl_ResetResult((Tcl_Interp *) iPtr);
    iPtr->resultCode = TCL_OK;
    code = Tcl_EvalStack (interp, STACK_CURRENT);
    return (code);
}

int Tcl_EvalKnownProcGlobal (interp,proc,data,argc,argv)
    Tcl_Interp *interp;
    Tcl_CmdProc proc;
    ClientData data;
    int argc;
    char **argv;
{
    register Interp *iPtr = (Interp *) interp;
    CallFrame *savedVarFramePtr;
    int code;

    savedVarFramePtr = iPtr -> varFramePtr;
    iPtr -> varFramePtr = NULL;
    code = Tcl_EvalKnownProc (interp, proc, data, argc, argv);
    iPtr -> varFramePtr = savedVarFramePtr;
    return (code);
}
 
#endif    /* STACK */ 
